/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "Camera.h"
#include <glm/gtx/rotate_vector.hpp>
#include <time.h>
#pragma warning(disable : 4996)
const float PI = float(atan(1.0) * 4.0);

CPhysicalCamera::CPhysicalCamera()
    : fPreTime(0), fSetTime(0), fTotalMoveTime(0)
    , bRoate_camera_flag(false), bMove_camera_flag(false)
    , fRotationAngleCube(0), fRotationAnglePyramid(0)
    , fCubeRotationSpeed(0), fPyramidRotationSpeed(0)
    , PIover180(0), iForw(0), iBack(0), iLeft(0)
    , iRight(0), flength(0)
{
    vEye = glm::vec3(0.0f, 0.0f, 0.0f);
    vView = glm::vec3(0.0f, 0.0, -1.0f);
    vUp = glm::vec3(0.0f, 1.0f, 0.0f);
    fSpeed = 25.0f;
    fSensitivity = 0.1f;
}

CPhysicalCamera::CPhysicalCamera(glm::vec3 a_vEye,
                                 glm::vec3 a_vView,
                                 glm::vec3 a_vUp,
                                 double a_fSpeed,
                                 double a_fSensitivity)
    : fPreTime(0), fSetTime(0), fTotalMoveTime(0)
    , bRoate_camera_flag(false), bMove_camera_flag(false)
    , fRotationAngleCube(0), fRotationAnglePyramid(0)
    , fCubeRotationSpeed(0), fPyramidRotationSpeed(0)
    , PIover180(0), iForw(0), iBack(0), iLeft(0)
    , iRight(0), flength(0)
{
    vEye = a_vEye; vView = a_vView; vUp = a_vUp;
    fSpeed = a_fSpeed;
    fSensitivity = a_fSensitivity;
}

float CPhysicalCamera::GetAngleX()
{
    glm::vec3 vDir = vView - vEye;
    vDir = glm::normalize(vDir);
    glm::vec3 vDir2 = vDir; vDir2.y = 0.0f;
    vDir2 = glm::normalize(vDir2);
    float fAngle =  acos(glm::dot(vDir2, vDir)) * (180.0f / PI);
    if (vDir.y < 0)
    {
        fAngle *= -1.0f;
    }
    return fAngle;
}

void CPhysicalCamera::rotateCamera(double horizontal, double vertical,
                                   double planar, double time)
{
    if (time != 0)
    {
        bRoate_camera_flag = true;
        fSetTime = time;
        glm::vec3 endView(horizontal, vertical, planar);
        glm::vec3 vRotate = endView - vView;
        vRotate /= time;
        vSpeed = vRotate;
    }
    else
    {
        float deltaX =  (float)(horizontal * fSensitivity);
        float deltaY =  (float)(vertical  * fSensitivity);
        glm::vec3 vCameraRotate = vView;
        if (deltaX != 0.0f)
        {
            vCameraRotate -= vEye;
            vCameraRotate = glm::rotate(vCameraRotate, deltaX,
                glm::vec3(0.0f, 1.0f, 0.0f));
            vCameraRotate += vEye;
        }
        if (deltaY != 0.0f)
        {
            glm::vec3 vAxis = glm::cross(vCameraRotate - vEye, vUp);
            vAxis = glm::normalize(vAxis);
            float fAngle = deltaY;
            float fNewAngle = fAngle + GetAngleX();
            if (fNewAngle > -89.80f && fNewAngle < 89.80f)
            {
                vCameraRotate -= vEye;
                vCameraRotate = glm::rotate(vCameraRotate, deltaY, vAxis);
                vCameraRotate += vEye;
            }
        }
        vView = vCameraRotate;
    }
}

void CPhysicalCamera::moveCamera(double x, double y, double z,
                                 double time)
{

    if (time != 0)
    {
        bMove_camera_flag = true;
        fSetTime = time;
        glm::vec3 endView(x, y, z);
        glm::vec3 vRotate = endView - vView;
        vRotate /= time ;
        vSpeed = vRotate;
    }
    else
    {
        fSetTime = 0.0;
        vView += glm::vec3(x, y, z);
        vEye += glm::vec3(x, y, z);
    }
}

void CPhysicalCamera::SetRotationAngle(double fRotateCube ,
                                       double fRotatePyramid ,
                                       double fCubeRotateSpeed,
                                       double fPyramidRotateSpeed)
{
    PIover180 = 3.1415f / 180.0f;
    fRotationAngleCube = fRotateCube;
    fRotationAnglePyramid = fRotatePyramid;
    fCubeRotationSpeed = fCubeRotateSpeed;
    fPyramidRotationSpeed = fPyramidRotateSpeed;
}

glm::mat4 CPhysicalCamera::Look()
{
    return glm::lookAt(vEye, vView, vUp);
}
glm::mat4 CPhysicalCamera::getModelViewMatrix(glm::mat4& lookat,glm::mat4& matrix)
{
    glm::mat4 modelViewMatrix;
    time_t tt = time(NULL);
    struct tm* t = localtime(&tt);
    int hour = t->tm_hour;
    int min = t->tm_min;
    int sec = t->tm_sec;
    int cTime = sec + min * 60 + hour * 3600;
    float fMoveTime = (float)(cTime - fPreTime);
    //glm::mat4 modelViewMatrix;
    if (fPreTime <= 0)
    {
        modelViewMatrix = lookat * matrix;
    }
    else
    {
        fTotalMoveTime += fMoveTime;
        if (fTotalMoveTime <= fSetTime)
        {
            if (bRoate_camera_flag == true)
            {
                vView += vSpeed * fMoveTime;
            }
            if (bMove_camera_flag == true)
            {
                vEye += vSpeed * fMoveTime;
                vView += vSpeed * fMoveTime;
            }
            modelViewMatrix = lookat * matrix;
        }
        else
        {
            modelViewMatrix = lookat * matrix;
        }
    }
    fPreTime = cTime;
    return modelViewMatrix;
}

void CPhysicalCamera::RotateObjectMouse(double horizontal, double vertical,
    double planar)
{
    fCubeRotationSpeed = 0.0f;
    fPyramidRotationSpeed = planar;
    float deltaX = (float)(horizontal * fSensitivity);
    float deltaY = (float)(vertical  * fSensitivity);
    if(deltaX != 0.0f)
    {
        fPyramidRotationSpeed -= deltaX * 3;
    }
    if(deltaY != 0.0f)
    {

        fCubeRotationSpeed -= deltaY * 3;
    }
    fRotationAngleCube += fCubeRotationSpeed;
    fRotationAnglePyramid += fPyramidRotationSpeed;

}