/*
 * This file is part of the LibreOffice project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
#include "RenderScene.h"
#include <glm/gtx/transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>
#include <cerrno>

RenderPrimitive::RenderPrimitive()
    : mVerterCount(0), mIndicesCount(0)
    , pMaterial(0), pNode(0)
    , mIndicesBuffer(0), mVertexBuffer(0)
    , mNormalBuffer(0), mTexCoordBuffer(0)
    , mJointBuffer(0), mWeightBuffer(0)
    , mIsPolygonSorted(false)
{
    mVertexBufferData = NULL;
    mIndiceBufferData = NULL;
    mSortedIndiceBufferData = NULL;
}

RenderPrimitive::~RenderPrimitive()
{
    if (NULL != mVertexBufferData)
    {
        free(mVertexBufferData);
    }
    if (NULL != mIndiceBufferData)
    {
        free(mIndiceBufferData);
    }
    if (NULL != mSortedIndiceBufferData)
    {
        free(mSortedIndiceBufferData);
    }
    glDeleteBuffers(1, &mVertexBuffer);
    glDeleteBuffers(1, &mNormalBuffer);
    glDeleteBuffers(1, &mTexCoordBuffer);
    glDeleteBuffers(1, &mIndicesBuffer);
}

Node* RenderPrimitive::getNode()
{
    return this->pNode;
}

void RenderPrimitive::setNode(Node* node)
{
    this->pNode = node;
}

Material* RenderPrimitive::getMaterial()
{
    return this->pMaterial;
}

void RenderPrimitive::setMaterial(Material* pMaterial)
{
    this->pMaterial = pMaterial;
}

unsigned int RenderPrimitive::getVerterCount()
{
    return this->mVerterCount;
}

void RenderPrimitive::setVerterCount(unsigned int count)
{
    this->mVerterCount = count;
}

unsigned int RenderPrimitive::getIndicesCount()
{
    return this->mIndicesCount;
}

void RenderPrimitive::setIndicesCount(unsigned int count)
{
    this->mIndicesCount = count;
}

DataType_E RenderPrimitive::getIndicesDataType()
{
    return this->mIndicesDataType;
}

void RenderPrimitive::setIndicesDataType(DataType_E type)
{
    this->mIndicesDataType = type;
}

void RenderPrimitive::copyIndiceBufferData(
                const char* srcBuf, unsigned int bufSize)
{
    if(NULL == mIndiceBufferData)
    {
        mIndiceBufferData = (char*)malloc(bufSize);
        memcpy(mIndiceBufferData,srcBuf,bufSize);
        mIndiceBufferLen = bufSize;
    }
    if (NULL == mSortedIndiceBufferData)
    {
        mSortedIndiceBufferData = (char*)malloc(bufSize);
    }
}
void RenderPrimitive::polyonSorting(glm::mat4& viewMatrix)
{
    mEyeSpaceVerts.resize(mVerterCount);
    glm::vec3* pVertices = (glm::vec3*)mVertexBufferData;
    for(size_t i=0; i<mVerterCount; ++i)
    {
        float z = pVertices[i].x*viewMatrix[2][0]
        + pVertices[i].y*viewMatrix[2][1]
        + pVertices[i].z*viewMatrix[2][2] + viewMatrix[2][3];
        mEyeSpaceVerts[i] = z;
    }
    sortIndices();
}
void RenderPrimitive::sortIndices()
{
    mPrimitiveZ.resize( mIndicesCount / 3 );
    if (mPrimitiveZ.empty())
    {
        mIsPolygonSorted = false;
        return;
    }
    if (DataType_UNSIGNED_SHORT == mIndicesDataType)
    {
        unsigned short* indexBuf = (unsigned short*) mIndiceBufferData;
        for(unsigned iz=0; iz < mIndicesCount / 3; ++iz)
        {
            mPrimitiveZ[iz].mZ = (float)(mEyeSpaceVerts[indexBuf[iz*3]]
            + mEyeSpaceVerts[indexBuf[iz*3+1]] + mEyeSpaceVerts[indexBuf[iz*3+2]]);
            mPrimitiveZ[iz].mPrimitiveIndex = iz;
        }
        std::sort( mPrimitiveZ.begin(), mPrimitiveZ.end(), Sorter_Back_To_Front() );
        unsigned short* sortedIndexBuf = (unsigned short *)mSortedIndiceBufferData;
        for(unsigned int i=0; i<mPrimitiveZ.size(); ++i)
        {
            sortedIndexBuf[i*3] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3];
            sortedIndexBuf[i*3+1] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3+1];
            sortedIndexBuf[i*3+2] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3+2];
        }
        mIsPolygonSorted = true;
    }else if (DataType_UNSIGNED_INT == mIndicesDataType)
    {
        unsigned int* indexBuf = (unsigned int*) mIndiceBufferData;
        for(unsigned iz=0; iz < mIndicesCount / 3; ++iz)
        {
            mPrimitiveZ[iz].mZ = (float)(mEyeSpaceVerts[indexBuf[iz*3]]
            + mEyeSpaceVerts[indexBuf[iz*3+1]] + mEyeSpaceVerts[indexBuf[iz*3+2]]);
            mPrimitiveZ[iz].mPrimitiveIndex = iz;
        }
        std::sort( mPrimitiveZ.begin(), mPrimitiveZ.end(), Sorter_Back_To_Front() );
        unsigned int* sortedIndexBuf = (unsigned int *)mSortedIndiceBufferData;
        for(unsigned int i=0; i<mPrimitiveZ.size(); ++i)
        {
            sortedIndexBuf[i*3] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3];
            sortedIndexBuf[i*3+1] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3+1];
            sortedIndexBuf[i*3+2] = indexBuf[ mPrimitiveZ[i].mPrimitiveIndex*3+2];
        }
        mIsPolygonSorted = true;
    }else
    {
        mIsPolygonSorted = false;
    }
}

void RenderPrimitive::bindSortedIndicesBuf()
{
    if (mIsPolygonSorted)
    {
        glDeleteBuffers(1, &mIndicesBuffer);
        glGenBuffers(1, &mIndicesBuffer);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mIndicesBuffer);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,
            mIndiceBufferLen,
            mSortedIndiceBufferData, GL_STATIC_DRAW);
    }
}

void RenderPrimitive::copyVertexBufferData(
                const char* srcBuf,unsigned int bufSize)
{
    if(NULL == mVertexBufferData)
    {
        mVertexBufferData = (char*)malloc(bufSize);
        memcpy(mVertexBufferData,srcBuf,bufSize);
    }
}

unsigned int RenderPrimitive::getVertexBuffer()
{
    return this->mVertexBuffer;
}

void RenderPrimitive::setVertexBuffer(unsigned int bufferId)
{
    this->mVertexBuffer = bufferId;
}

unsigned int RenderPrimitive::getNormalBuffer()
{
    return this->mNormalBuffer;
}

void RenderPrimitive::setNormalBuffer(unsigned int bufferId)
{
    this->mNormalBuffer = bufferId;
}

unsigned int RenderPrimitive::getTexCoordBuffer()
{
    return this->mTexCoordBuffer;
}

void RenderPrimitive::setTexCoordBuffer(unsigned int bufferId)
{
    this->mTexCoordBuffer = bufferId;
}

unsigned int RenderPrimitive::getJointBuffer()
{
    return this->mJointBuffer;
}

void RenderPrimitive::setJointBuffer(unsigned int bufferId)
{
    this->mJointBuffer = bufferId;
}

unsigned int RenderPrimitive::getWeightBuffer()
{
    return this->mWeightBuffer;
}

void RenderPrimitive::setWeightBuffer(unsigned int bufferId)
{
    this->mWeightBuffer = bufferId;
}

unsigned int RenderPrimitive::getIndicesBuffer()
{
    return this->mIndicesBuffer;
}

void RenderPrimitive::setIndicesBuffer(unsigned int bufferId)
{
    this->mIndicesBuffer = bufferId;
}

RenderWithFBO::RenderWithFBO()
{
    mFboProId = 0;
    mFboId = 0;
    mRboId = 0;
    mTexId = 0;
    mShotFboId = 0;
    mShotRboId = 0;
    mShotTexId = 0;
    mInverseFboId = 0;
    mInverseRboId = 0;
    mInverseTexId = 0;
    mVertexBuf = 0;
    mTexCoordBuf = 0;
}

void RenderWithFBO::inverseBitMap(int width, int height)
{
    GLuint proId = loadFboShader(INVERSEVERT, INVERSEFRAG);
    GLuint texCoordBuf = 0;
    GLuint vertexBuf = 0;
    GLfloat coordVertices[] =
    {
        0.0f, 1.0f,
        1.0f, 1.0f,
        1.0f, 0.0f,
        0.0f, 0.0f,
    };
    GLfloat squareVertices[] =
    {
        -1.0f, -1.0f, -1.0,
        1.0f, -1.0f, -1.0,
        1.0f,  1.0f, -1.0,
        -1.0f,  1.0f, -1.0
    };
    setBufferForFbo(texCoordBuf, vertexBuf, coordVertices,
                    sizeof(coordVertices), squareVertices,
                    sizeof(squareVertices));
    createAndBindInverseFBO(width, height);
    glViewport(0, 0, width, height);
    inverseTexture(proId, texCoordBuf, vertexBuf);
}

GLuint RenderWithFBO::loadFboShader(const char* vShader,
                                    const char* fShader)
{
    ShaderProgram shaderPro;
    GLuint progId = glCreateProgram();
    shaderPro.loadShader(progId, vShader, strlen(vShader), GL_VERTEX_SHADER);
    shaderPro.loadShader(progId, fShader, strlen(fShader), GL_FRAGMENT_SHADER);
    return progId;
}

void RenderWithFBO::setBufferForFbo(GLuint& texCoordBuf, GLuint& vertexBuf,
                                    GLfloat* pCoord, GLuint numCoord,
                                    GLfloat* pSquare, GLuint numSquare)
{
    glGenBuffers(1, &texCoordBuf);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glBufferData(GL_ARRAY_BUFFER, numCoord, pCoord, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glGenBuffers(1, &vertexBuf);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glBufferData(GL_ARRAY_BUFFER, numSquare, pSquare, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void RenderWithFBO::createAndBindInverseFBO(int width, int height)
{
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    createRenderObj(width, height, mInverseRboId);
    createTextureObj(width, height, mInverseTexId);
    createFrameBufferObj(mInverseFboId, mInverseTexId, mInverseRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mInverseFboId);
}

void RenderWithFBO::createAndBindBitMapFBO(int width, int height)
{
    createRenderObj(width, height, mShotRboId);
    createTextureObj(width, height, mShotTexId);
    createFrameBufferObj(mShotFboId, mShotTexId, mShotRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mShotFboId);
}

void RenderWithFBO::inverseTexture(GLuint proId, GLuint texCoordBuf,
                                   GLuint vertexBuf)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    GLuint texCoordId = glGetAttribLocation(proId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(proId, "vPosition");
    GLuint textureId  = glGetUniformLocation(proId, "RenderTex");

    glUseProgram(proId);
    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glVertexAttribPointer(vertexId, 3, GL_FLOAT, GL_FALSE,
                          0, (void*)0);

    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE,
                          0, (void*)0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, mShotTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(vertexId);
    glDisableVertexAttribArray(texCoordId);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glUseProgram(0);
}

void RenderWithFBO::createRenderObj(int width, int height, GLuint& rboId)
{
    glGenRenderbuffers(1, &rboId);
    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
}

void RenderWithFBO::createTextureObj(int width, int height, GLuint& texId)
{
    glGenTextures(1, &texId);
    glBindTexture(GL_TEXTURE_2D, texId);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB,
                 GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
}

bool RenderWithFBO::createFrameBufferObj(GLuint& fboId,
        GLuint texId, GLuint rboId)
{
    glGenFramebuffers(1, &fboId);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, fboId);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, texId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                           GL_TEXTURE_2D, texId, 0);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);

    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                              GL_RENDERBUFFER, rboId);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return false;
    }
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    return true;
}

int RenderWithFBO::renderFboTexture()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(mFboProId);

    GLuint texCoordId = glGetAttribLocation(mFboProId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(mFboProId, "vPosition");
    GLuint textureId  = glGetUniformLocation(mFboProId, "RenderTex");

    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, mVertexBuf);
    glVertexAttribPointer(vertexId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, mTexCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glBindTexture(GL_TEXTURE_2D, mTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(texCoordId);
    glDisableVertexAttribArray(vertexId);
    return 0;
}

void RenderWithFBO::createAndBindFbo(int width, int height)
{
    if (0 == mFboId)
    {
        mFboProId = loadFboShader(FBOVERT, FBOFRAG);
        createRenderObj(width, height, mRboId);
        createTextureObj(width, height, mTexId);
        createFrameBufferObj(mFboId, mTexId, mRboId);

        GLfloat coordVertices[] =
        {
            0.0f, 0.0f,
            1.0f, 0.0f,
            1.0f, 1.0f,
            0.0f, 1.0f,
        };
        GLfloat squareVertices[] =
        {
            -1.0f, -1.0f,
            1.0f, -1.0f,
            1.0f,  1.0f,
            -1.0f,  1.0f,
        };
        setBufferForFbo(mTexCoordBuf, mVertexBuf, coordVertices,
            sizeof(coordVertices), squareVertices, sizeof(squareVertices));
    }
}

void RenderWithFBO::releaseFbo()
{
    if (0 != mFboId)
    {
        glDeleteFramebuffers(1, &mFboId);
        glDeleteRenderbuffers(1, &mRboId);
        glDeleteTextures(1, &mTexId);
        mFboId = 0;
     
    }
    if(0 != mShotFboId)
    {
        glDeleteFramebuffers(1, &mShotFboId);
        glDeleteRenderbuffers(1, &mShotRboId);
        glDeleteTextures(1, &mShotTexId);
        mShotFboId = 0;
    }
}



ShaderProgram RenderScene::mShaderProgram = ShaderProgram();

RenderScene::RenderScene()
    : pScene(0), pLight(0), pCamera(0)
    , mCurrentTime(0), mLastPlaying(0), mDuration(0)
    , mAnimationPlay(false), mAnimationLoop(true)
    , mEnableRotation(false), mRotationDuration(0)
    , mRotationNow(true)
{
#if ENABLE_FPS
    pFPSCounter = NULL;
#endif
#if ENABLE_TIMER
    pTimer = new Timer();
#endif
}

RenderScene::~RenderScene()
{
#if ENABLE_FPS
    if (pFPSCounter != NULL)
    {
        delete pFPSCounter;
    }
#endif
#if ENABLE_TIMER
    if(pTimer != NULL)
    {
        delete pTimer;
    }
#endif
    if (pScene->getLight() && pLight != NULL)
    {
        delete pLight;
    }
    if (pCamera != NULL)
    {
        delete pCamera;
    }
    unsigned int size = mPrimitiveVec.size();
    for (unsigned int i = 0; i < size; ++i)
    {
        if (mPrimitiveVec[i] != NULL)
        {
             delete mPrimitiveVec[i];
        }
    }
    mPrimitiveVec.clear();
    mBindBufferMap.clear();
    if (pScene != NULL)
    {
        delete pScene;
    }
}

Scene* RenderScene::loadScene(const std::string& directory,
                              const std::string& file)
{
    Scene* pScene = new Scene();
    mLoadJson.setJsonInfo(directory, file);
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }

    if (!mLoadJson.parseScene(pScene))
    {
        return 0;
    }
    return pScene;
}

Scene* RenderScene::loadScene(glTFHandle* handle)
{
    Scene* pScene = new Scene();
    pScene->setGltfHandle(handle);
    try
    {
#if ENABLE_TIMER
        static int loadSceneCounter;
        pTimer->startTimer("loadScene->parseScene", loadSceneCounter);
        if (!mLoadJson.parseScene(pScene))
        {
            return 0;
        }
        pTimer->stopTimer();
        loadSceneCounter++;
#else
        if (!mLoadJson.parseScene(pScene))
        {
            return 0;
        }
#endif
    }
    catch (boost::property_tree::ptree_error& e)
    {
        fprintf(stderr, "%s\n", e.what());
        delete pScene;
        return 0;
    }
    catch (boost::exception const&)
    {
        fprintf(stderr, "%s\n", "unknown boost error");
        delete pScene;
        return 0;
    }
    
    return pScene;
}

glTFHandle* RenderScene::initScene(const glTFFile* jsonfile)
{
    if (0 == jsonfile)
    {
        return 0;
    }
    string tmpStr(jsonfile->buffer, jsonfile->size);
    mLoadJson.setJsonInfo(tmpStr);
#if ENABLE_TIMER
    /* Counter on each section is to find the the no of times the control
     * enters this sections and we select the model name based on the
     * counter value. */
    static int initCounter;
    pTimer->startTimer("initScene->parseJsonFile", initCounter);
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }
    pTimer->stopTimer();
    initCounter++;
#else
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }
#endif
    string tmpFileName(jsonfile->filename);
    return mLoadJson.getFileNameInJson(tmpFileName);
}

unsigned int RenderScene::bindAttribute(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

unsigned int RenderScene::bindIndices(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

void RenderScene::constructMesh(const std::string& meshName, Node* pNode)
{
    const Mesh* pMesh = pScene->findMesh(meshName);
    unsigned int size = pMesh->getPrimitiveVecSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        constructPrimitive(pMesh->getPrimitiveVec(i), pNode);
    }
}

void RenderScene::constructPrimitive(const Primitives* pPrimitive, Node* pNode)
{
    Material* pMaterial = pScene->findMaterial(
                            pPrimitive->getMaterialIndex());
    if (0 == pMaterial)
    {
        return;
    }
    RenderPrimitive* pRenderPrimitive = new RenderPrimitive();
    bindAttributeBuffer(pPrimitive, pRenderPrimitive);
    pRenderPrimitive->setMaterial(pMaterial);
    pRenderPrimitive->setNode(pNode);
    pushPrimitiveVec(pRenderPrimitive);
}

void RenderScene::bindAttributeBuffer(const Primitives* pPrimitive,
                                      RenderPrimitive* pRP)
{
    const Attribute* pAttr;
    BindBufferInfo bindBufferInfo;
    std::map<std::string, BindBufferInfo>::const_iterator it;

    const std::string& posId = pPrimitive->getAttributeIndex("POSITION");
    it = mBindBufferMap.find(posId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(posId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            bindBufferInfo.mDataCount = pAttr->getDataCount();
            bindBufferInfo.mBufferLen = pAttr->getDataCount()
                                        * pAttr->getByteStride();
            bindBufferInfo.mSrcData = pAttr->getAttributeData();
            pRP->setVertexBuffer(bindBufferInfo.mBufferId);
            pRP->setVerterCount(bindBufferInfo.mDataCount);
#if POLYGONSORTING
            pRP->copyVertexBufferData(bindBufferInfo.mSrcData,
                                      bindBufferInfo.mBufferLen);
#endif
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (posId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setVertexBuffer(it->second.mBufferId);
        pRP->setVerterCount(it->second.mDataCount);
#if POLYGONSORTING
        pRP->copyVertexBufferData(it->second.mSrcData,
                                  it->second.mBufferLen);
#endif
    }

    const std::string& NorId = pPrimitive->getAttributeIndex("NORMAL");
    it = mBindBufferMap.find(NorId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(NorId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setNormalBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (NorId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setNormalBuffer(it->second.mBufferId);
    }

    const std::string& TexId = pPrimitive->getAttributeIndex("TEXCOORD_0");
    it = mBindBufferMap.find(TexId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(TexId);
        if (0 != pAttr)
        {
            float* pData = (float*)pAttr->getAttributeData();
            unsigned int size = pAttr->getDataCount();
            for (unsigned int i = 0; i < size; i++)
            {
                pData[i*2+1] = 1 - pData[i*2+1];
            }
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setTexCoordBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (TexId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setTexCoordBuffer(it->second.mBufferId);
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("JOINT"));
    if (0 != pAttr)
    {
        pRP->setJointBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("WEIGHT"));
    if (0 != pAttr)
    {
        pRP->setWeightBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getIndicesIndex());
    if (0 != pAttr)
    {
        pRP->setIndicesBuffer(bindIndices(pAttr));
        pRP->setIndicesCount(pAttr->getDataCount());
        pRP->setIndicesDataType(pAttr->getDataType());
#if POLYGONSORTING
        pRP->copyIndiceBufferData(pAttr->getAttributeData(),
            pAttr->getDataCount() * pAttr->getByteStride());
#endif
    }
}

void RenderScene::pushPrimitiveVec(RenderPrimitive* pPrim)
{
    (this->mPrimitiveVec).push_back(pPrim);
}

RenderPrimitive* RenderScene::getPrimitiveVec(unsigned int index)
{
    if (this->mPrimitiveVec.size() <= index)
    {
        return 0;
    }
    return this->mPrimitiveVec[index];
}

unsigned int RenderScene::getPrimitiveVecSize()
{
    return this->mPrimitiveVec.size();
}

const glm::mat4 RenderScene::calcProjectionMatrix()
{
    glm::mat4 projection;
    if (!pScene->getFlagCamera())
    {
        projection = glm::perspective(37.8492f, 1.5f, 1.0f, 500000.0f);
    }
    else
    {
        projection = glm::perspective(cCamera->getXFov(),
                                      cCamera->getAspectRatio(),
                                      cCamera->getNear(),
                                      cCamera->getFar());
    }
    return projection;
}

void RenderScene::getCameraIndex(Node* rootNode)
{
    for (unsigned int i = 0; i < rootNode->getChildNodeSize(); i++)
    {

        Node* pNode = rootNode->getChildNode(i);
        if (pNode->getCameraIndex() != "")
        {
            vCameraIndex.push_back(pNode->getCameraIndex());
        }
        getCameraIndex(pNode);
    }
    return;
}
void RenderScene::initOpengl()
{
    glClearColor(0.5f, 0.5f, 0.5f, 0.5f);
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glDepthFunc(GL_LESS);
}

void RenderScene::createDefaultCamera()
{
    glm::vec3 vDelta = pScene->getVertexMax() - pScene->getVertexMin();
    glm::vec3 vHalf =  pScene->getVertexMin() + (vDelta / 2.0f);

    float fPos = sqrt(vDelta.x * vDelta.x
                      + vDelta.y * vDelta.y
                      + vDelta.z * vDelta.z);
    if (fPos >= 3000)
    {
        pCamera = new CPhysicalCamera(
            glm::vec3(vHalf.x, -fPos, 0.1f*fPos),
            vHalf,
            glm::vec3(0.0f, 1.0f, 0.0f));
    }
    else
    {
        pCamera = new CPhysicalCamera(
            glm::vec3(vHalf.x, 1*fPos, -2.5f*fPos),
            vHalf,
            glm::vec3(0.0f, 1.0f, 0.0f));
    }
    pCamera->flength = fPos;
    pCamera->vModelCenterPos = vHalf;
    pCamera->SetRotationAngle();
}

int RenderScene::initRender(glTFHandle* handle)
{
    pScene = loadScene(handle);
    if (0 == pScene)
    {
        return LIBGLTF_SCENE_NONE;
    }

    initOpengl();
    createDefaultCamera();

    getCameraIndex(pScene->getRootNode());
    std::vector<std::string>::iterator it;
    it = vCameraIndex.begin();
    cCamera = pScene->findCamera(*it);

    Node* pNode;
    unsigned int nSize = pScene->getNodeSize();
    for (unsigned int i = 0; i < nSize; ++i)
    {
        pNode = pScene->getNode(i);
        if (!pNode->hasMesh())
        {
            continue;
        }
        unsigned int mSize = pNode->getMeshIndexSize();
        for (unsigned int j = 0; j < mSize; ++j)
        {
            constructMesh(pNode->getMeshIndex(j), pNode);
        }
    }

    this->mDuration = pScene->getDuration();

    return 0;
}

void RenderScene::initFPS(unsigned char* fontBuf,unsigned int bufSize)
{
#if ENABLE_FPS
    if (NULL != pFPSCounter)
    {
        delete pFPSCounter;
    }
    pFPSCounter = new FPSCounter(fontBuf,bufSize,&mShaderProgram);
#endif
}

void RenderScene::renderPrimitive(RenderPrimitive* pPrimitive,
                                  unsigned int progId)
{
    upLoadMatrixInfo(progId, pPrimitive);

#if POLYGONSORTING
    pPrimitive->bindSortedIndicesBuf();
#endif

#if ENABLE_TIMER
    static int upLoadUniformCounter;
    pTimer->startTimer("RenderPrimitive->upLoadUniform", upLoadUniformCounter);
    upLoadUniform(progId, pPrimitive);
    pTimer->stopTimer();
    upLoadUniformCounter++;
#else
    upLoadUniform(progId, pPrimitive);
#endif

#if ENABLE_TIMER
    static int upLoadAttributeCounter;
    pTimer->startTimer("RenderPrimitive->upLoadAttribute", upLoadAttributeCounter);
    upLoadAttribute(progId, pPrimitive);
    pTimer->stopTimer();
    upLoadAttributeCounter++;
#else
    upLoadAttribute(progId, pPrimitive);
#endif

#if ENABLE_TIMER
    static int upLoadAnimationCounter;
    pTimer->startTimer("RenderPrimitive->upLoadAnimation", upLoadAnimationCounter);
    if (pScene->getSkinSize() > 0)
    {
        upLoadAnimation(progId, pPrimitive);
    }
    pTimer->stopTimer();
    upLoadAnimationCounter++;
#else
    if (pScene->getSkinSize() > 0)
    {
        upLoadAnimation(progId, pPrimitive);
    }
#endif

#if ENABLE_TIMER
    static int drawTriangleCounter;
    pTimer->startTimer("RenderPrimitive->drawTriangles", drawTriangleCounter);
    drawTriangle(pPrimitive);
    pTimer->stopTimer();
    drawTriangleCounter++;
#else
    drawTriangle(pPrimitive);
#endif
}

void RenderScene::upLoadMatrixInfo(unsigned int progId,
                                   RenderPrimitive* pPrimitive)
{
    Node* pNode = pPrimitive->getNode();
    glm::mat4 globalMatrix = pNode->getGlobalMatrix();
    glm::mat4 temp;
    if (pScene->getFlagCamera())
    {
        temp = cCamera->getCameraNode()->getGlobalMatrix();
        temp = glm::inverse(temp);
    } 
    else
    {
        temp = pCamera->Look();
    }

    glm::mat4 modelViewMatrix = pCamera->getModelViewMatrix(temp,globalMatrix);
    glm::mat3 normalMatrix(globalMatrix);
    glm::mat3 normalInverseMatrix =
                glm::mat3(temp) *
                glm::transpose(glm::inverse(normalMatrix));
    const glm::mat4 project = calcProjectionMatrix();

#if ENABLE_TIMER
    static int upLoadMatrixInfoCounter;
    pTimer->startTimer("RenderPrimitive->upLoadMatrixInfo",
                        upLoadMatrixInfoCounter);
    mShaderProgram.setUniform(progId, "u_modelViewMatrix", modelViewMatrix);
    mShaderProgram.setUniform(progId, "u_normalMatrix", normalInverseMatrix);
    mShaderProgram.setUniform(progId, "u_projectionMatrix", project);
    mShaderProgram.setUniform(progId, "M", (globalMatrix));
    mShaderProgram.setUniform(progId, "V", (pCamera->Look()));
    pTimer->stopTimer();
    upLoadMatrixInfoCounter++;
#else
    mShaderProgram.setUniform(progId, "u_modelViewMatrix", modelViewMatrix);
    mShaderProgram.setUniform(progId, "u_normalMatrix", normalInverseMatrix);
    mShaderProgram.setUniform(progId, "u_projectionMatrix", project);
    mShaderProgram.setUniform(progId, "M", (globalMatrix));
    mShaderProgram.setUniform(progId, "V", (temp));
#endif
}

bool RenderScene::upLoadTechProperty()
{
    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_TRUE);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    return true;
}

void RenderScene::upLoadTechInfo(unsigned int progId, Technique* pTech)
{
    std::vector<TechLight*> vecLight = pTech->popTechLight();
    std::vector<TechLight*>::iterator it = vecLight.begin();
    std::vector<TechLight*>::iterator itL = vecLight.end();
    for (; it != itL; ++it)
    {
        TechLight *pTechLight = *it;
        std::string strSource = pTechLight->getSource();
        std::string strName = pTechLight->getTechniLightName();
        if (strSource.empty())
        {
            mShaderProgram.setUniform(progId, 
                strName.c_str(),
                pTechLight->getValue());
        }
        else
        {
            Node *pNode = pScene->findLightNodeMap(strSource);
            mShaderProgram.setUniform(progId,
                strName.c_str(),
                pNode->getGlobalMatrix());
        }
    }
}

void RenderScene::upLoadAnimation(unsigned int progId,
                                  RenderPrimitive* pPrimitive)
{
    Node* pNode = pPrimitive->getNode();
    const std::string& skinId = pNode->getSkinIndex();
    if (skinId.empty())
        return;
    Skin* pSkin = 0;
    unsigned int size = pScene->getSkinSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        pSkin = pScene->getSkin(i);
        if (pSkin->getSkinName() != skinId)
            continue;
        break;
    }
    glm::mat4* pBindMatrix = pSkin->getBindMatrix();
    glm::mat4* pTempMatrix = new glm::mat4[pSkin->getBindMatrixCount()];
    memcpy(pTempMatrix, pBindMatrix,
           sizeof(glm::mat4)*pSkin->getBindMatrixCount());

    unsigned int jointCount = pSkin->getBoneIdSize();
    for (unsigned int i = 0; i < jointCount; i++)
    {
        pNode = pScene->findBoneNode(pSkin->getBoneId(i));
        pTempMatrix[i] = pNode->getGlobalMatrix() * pTempMatrix[i];
    }

    int iLoc = glGetUniformLocation(progId, "u_jointMat");
    glUniformMatrix4fv(iLoc, pSkin->getBindMatrixCount(), false,
                       (GLfloat*)pTempMatrix);
    delete[] pTempMatrix;
}

void RenderScene::updateAnimInfo(Node* pNode)
{
    Animation* pAnimation = pScene->findAnimation(pNode->getNodeName());
    if (0 != pAnimation)
    {
        double time = fmod(mCurrentTime, pAnimation->getDuration());
        AnimChannel* pChannel = pAnimation->getRotateChl();
        if (0 != pChannel)
        {
            pNode->setRotate(pChannel->getTimeValue(time));
        }
        pChannel = pAnimation->getTransChl();
        if (0 != pChannel)
        {
            pNode->setTranslate(pChannel->getTimeValue(time));
        }
        pChannel = pAnimation->getScaleChl();
        if (0 != pChannel)
        {
            pNode->setScale(pChannel->getTimeValue(time));
        }

        glm::mat4 localMatrix = pNode->getTranslate()
                              * pNode->getRotate()
                              * pNode->getScale();
        pNode->setLocalMatrix(localMatrix);
    }
}

void RenderScene::updateNodeMatrix(Node* pNode, const glm::mat4& matrix)
{
    if (this->mAnimationPlay)
    {
        updateAnimInfo(pNode);
    }
    if (this->mAnimationPlay || mRotationNow)
    {
        glm::mat4 globalMatrix = matrix * pNode->getLocalMatrix();
        pNode->setGlobalMatrix(globalMatrix);
        unsigned int nSize = pNode->getChildNodeSize();
        for (unsigned int i = 0; i < nSize; ++i)
        {
            updateNodeMatrix(pNode->getChildNode(i), globalMatrix);
        }
    }
    return;
}

void RenderScene::upLoadUniform(unsigned int progId,
                                RenderPrimitive* pPrimitive)
{
    int textureId = 0;
    Material* pMaterial = pPrimitive->getMaterial();
    unsigned int size = pMaterial->getMaterialProperSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        const MaterialProperty* pProper = pMaterial->getMaterialProper(i);
        const std::string& properName = pProper->getPropertyName();
        switch (pProper->getDataType())
        {
            case DataType_FLOAT:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (float*)(pProper->
                                                   getPropertyData()));
                break;
            case DataType_FLOAT_VEC2:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec2*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_VEC3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec3*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_VEC4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::vec4*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_MAT3:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::mat3*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_FLOAT_MAT4:
                mShaderProgram.setUniform(progId, properName.c_str(),
                                          (glm::mat4*)(pProper->
                                                       getPropertyData()));
                break;
            case DataType_SAMPLER_2D:
                {
#if LOAD_ONCE
                    if (mCurrentImage == pProper->getImagePath())
                    {
                        break;
                    }
                    mCurrentImage = pProper->getImagePath();
#endif 
                    mShaderProgram.setUniform(progId, properName.c_str(),
                        textureId);
                    Texture* oneTexture =
                        pScene->findTexture(pProper->getImagePath());
                    oneTexture->bindTexture(textureId);
                    textureId++;
                }
                break;
            default:
                break;
        }
    }
    return;
}

void RenderScene::upLoadAttribute(unsigned int progId,
                                  RenderPrimitive* pPrimitive)
{
    int tmpId = glGetAttribLocation(progId, "a_position");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getVertexBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_normal");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getNormalBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_texcoord0");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getTexCoordBuffer());
        glVertexAttribPointer(tmpId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_weight");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getWeightBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }

    tmpId = glGetAttribLocation(progId, "a_joint");
    if (-1 != tmpId)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getJointBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    return;
}

void RenderScene::drawTriangle(RenderPrimitive* pPrimitive)
{
    if (pPrimitive->getIndicesCount() > 0)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, pPrimitive->getIndicesBuffer());
        glDrawElements(GL_TRIANGLES, pPrimitive->getIndicesCount(),
                       pPrimitive->getIndicesDataType(), 0);
    }
    else
    {
        glDrawArrays(GL_TRIANGLES, 0, pPrimitive->getVerterCount());
    }
}

void RenderScene::prepareRender(void* lpParam)
{
#if ENABLE_FPS
    if(pFPSCounter != NULL)
    {
        pFPSCounter->timeStamp();
    }
#endif
    viewport = *((glTFViewport*)lpParam);
#if ENABLE_TIMER
    static int prepareRenderCounter;
    pTimer->startTimer("repareRender", prepareRenderCounter);
    fbo.createAndBindFbo(viewport.width * SSAA, viewport.height * SSAA);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo.mFboId);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(viewport.x, viewport.y, viewport.width * SSAA, viewport.height * SSAA);
    pTimer->stopTimer();
    prepareRenderCounter++;
#else
    fbo.createAndBindFbo(viewport.width * SSAA, viewport.height * SSAA);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo.mFboId);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(viewport.x, viewport.y, viewport.width * SSAA, viewport.height * SSAA);
#endif
    return;
}
double RenderScene::getLastFrameDuration()
{
    double durationTime = 0;
    double currentTime = 0;
#ifdef _WIN32
    LARGE_INTEGER f;
    QueryPerformanceFrequency(&f);
    _LARGE_INTEGER tmp;
    QueryPerformanceCounter(&tmp);
    currentTime = (((double)tmp.QuadPart)/(double)f.QuadPart)*1000;
#else
#ifdef _linux_
    struct timeval tmp;
    gettimeofday(&tmp, NULL);
    currentTime = tmp.tv_sec*1000 + (tmp.tv_usec / 1000);
#else
    currentTime = clock() / double(CLOCKS_PER_SEC);
#endif
#endif
    if (mRotationLastTime != 0)
    {
        durationTime = currentTime - mRotationLastTime;
    }
    mRotationLastTime = currentTime;
    return durationTime;
}

void RenderScene::updateRotationStatus()
{
    mRotationDuration += getLastFrameDuration();
    if (mRotationDuration > ((double)CLOCKS_PER_SEC / 30))
    {
        mRotationNow = true;
        mRotationDuration = 0;
    }
    else
    {
        mRotationNow = false;
    }
}
void RenderScene::render()
{
    mCurrentTime += clock() / double(CLOCKS_PER_SEC) - mLastPlaying;
    updateRotationStatus();
    Node* pRootNode = pScene->getRootNode();
    glm::mat4 rootMatrix = pRootNode->getGlobalMatrix();
    glm::mat4 tmpMat4_X = glm::mat4(1.0);
    glm::mat4 tmpMat4_Y = glm::mat4(1.0);
    tmpMat4_X = glm::rotate((float)pCamera->fCubeRotationSpeed,
                            glm::vec3(1.0f, 0.0f, 0.0f));
    tmpMat4_Y = glm::rotate((float)pCamera->fPyramidRotationSpeed,
                            glm::vec3(.0f, 1.0f, 0.0f));
    if (mEnableRotation && mRotationNow)
        tmpMat4_Y = glm::rotate(0.3f, glm::vec3(0.0f, 1.0f, 0.0f));
    rootMatrix = tmpMat4_X * tmpMat4_Y * rootMatrix;
    updateNodeMatrix(pScene->getRootNode(), rootMatrix);
#if POLYGONSORTING
    updatePolygonSorting();
#endif
    unsigned int size = pScene->getTechSize();
#if ENABLE_TIMER
    static int renderCounter;
    pTimer->startTimer("render", renderCounter);
    int skey = pTimer->startKey;
    ++(pTimer->endKey);
    upLoadTechProperty();
    for (unsigned int i = 0; i < size; ++i)
    {
        renderShader(pScene->getTechnique(i));
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    int eKey = pTimer->endKey;
    pTimer->endKey = --skey;
    pTimer->stopTimer();
    pTimer->endKey = eKey;
    renderCounter++;
#else
    upLoadTechProperty();
    for (unsigned int i = 0; i < size; ++i)
    {
        renderShader(pScene->getTechnique(i));
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
#endif
    mLastPlaying = mCurrentTime;
}
void RenderScene::primitivePolygonSorting(int i)
{
    RenderPrimitive* pPrimitive = getPrimitiveVec(i);
    Node* pNode = pPrimitive->getNode();
    glm::mat4 globalMatrix = pNode->getGlobalMatrix();
    glm::mat4 viewMatrix;
    if (pScene->getFlagCamera())
    {
        viewMatrix = cCamera->getCameraNode()->getGlobalMatrix();
        viewMatrix = glm::inverse(viewMatrix);
    } 
    else
    {
        viewMatrix = pCamera->Look();
    }
    glm::mat4 modelViewMatrix = pCamera->getModelViewMatrix(viewMatrix,globalMatrix);
    pPrimitive->polyonSorting(modelViewMatrix);
}
void RenderScene::updatePolygonSorting()
{
    int size = getPrimitiveVecSize();
#if OPENMPENABLE
    int loopSize = size/WAVESIZE;
    int remainder = size - loopSize*WAVESIZE;
    for(int i = 0;i < loopSize; ++i)
    {
#pragma omp parallel for
        for (int j = 0;j<WAVESIZE;++j)
        {
            primitivePolygonSorting(i*WAVESIZE+j);
        }
    }
    for (int i = 0;i < remainder;++i)
    {
        primitivePolygonSorting(loopSize*WAVESIZE+i);
    }

#else
    for (int i = 0; i < size; ++i)
    {
        primitivePolygonSorting(i);
    }
#endif
}

void RenderScene::renderShader(Technique* pTechnique)
{
    if (!pTechnique->getProgramState())
    {
        return;
    }

    unsigned int progId = pTechnique->getProgramId();
#if ENABLE_TIMER
    static int upLoadTechInfoCounter;
    pTimer->startTimer("RenderShader->upLoadTechInfo", upLoadTechInfoCounter);
    pTechnique->useTechnique();
    upLoadTechInfo(progId, pTechnique);
    pTimer->stopTimer();
    upLoadTechInfoCounter++;
#else
    pTechnique->useTechnique();
    upLoadTechInfo(progId, pTechnique);
#endif

    const std::string& techId = pTechnique->getTechId();
    unsigned int size = getPrimitiveVecSize();
    for (unsigned int i = 0; i < size; ++i)
    {
        RenderPrimitive* pPrimitive = getPrimitiveVec(i);
        const Material* pMaterial = pPrimitive->getMaterial();
        if (pMaterial->getTechniqueId() == techId)
        {
            renderPrimitive(pPrimitive, progId);
        }
    }
}

void RenderScene::completeRender()
{
#if ENABLE_FPS
    if (pFPSCounter != NULL)
    {
        pFPSCounter->printFPS(&viewport);
    }
#endif
#if ENABLE_TIMER
    static int completeRenderCounter;
    pTimer->startTimer("completeRender",completeRenderCounter);
    glViewport(viewport.x, viewport.y, viewport.width, viewport.height);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    fbo.renderFboTexture();
    pTimer->stopTimer();
    completeRenderCounter++;
#else
    glViewport(viewport.x, viewport.y, viewport.width, viewport.height);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    fbo.renderFboTexture();
#endif
    return;
}

void RenderScene::releaseRender()
{
    mLoadJson.releaseFileName();
    fbo.releaseFbo();
    return;
}

CPhysicalCamera* RenderScene::getCamera()
{
    return pCamera;
}

void RenderScene::prepareRenderBitmap(glTFViewport* pViewport)
{
    fbo.createAndBindBitMapFBO(pViewport->width, pViewport->height);
    glViewport(pViewport->x, pViewport->y, pViewport->width, pViewport->height);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
}

void RenderScene::renderBitmap()
{
    render();
}

void RenderScene::completeRenderBitmap(glTFViewport* pViewport,
                                       unsigned char* buffer, GLenum format)
{
    glDisable(GL_DEPTH_TEST);
    fbo.inverseBitMap(pViewport->width, pViewport->height);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return;
    }
    glReadPixels(0, 0, pViewport->width, pViewport->height, format,
                 GL_UNSIGNED_BYTE, buffer);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glDeleteFramebuffers(1, &fbo.mInverseFboId);
    glDeleteRenderbuffers(1, &fbo.mInverseRboId);
    glDeleteTextures(1, &fbo.mInverseTexId);
}

void RenderScene::getCameraPos(glm::vec3* Eye, glm::vec3* view,
                               glm::vec3* up)
{
    *Eye =  pCamera->vEye;
    *view = pCamera->vView;
    *up = pCamera->vUp;
}

glm::vec3* RenderScene::getModelCenterPos()
{
    return &pCamera->vModelCenterPos;
}

void RenderScene::EnableRotation()
{
    mEnableRotation = true;
}

void RenderScene::DisableRotation()
{
    mEnableRotation = false;
}

double RenderScene::getModelSize()
{
    return pCamera->flength;
}

void RenderScene::startAnimation()
{
    this->mAnimationPlay = true;
}

void RenderScene::stopAnimation()
{
    this->mAnimationPlay = false;
}

void RenderScene::setAnimTime(double time)
{
    this->mLastPlaying = clock() / double(CLOCKS_PER_SEC) - time;
}

double RenderScene::getAnimTime()
{
    errno = 0;
    double time = fmod(this->mCurrentTime, this->mDuration);
    return errno == EDOM ? 0.0 : time;
}

void RenderScene::setAnimLoop(int loop)
{
    if (0 != loop)
    {
        this->mAnimationLoop = true;
    }
    else
    {
        this->mAnimationLoop = false;
    }
}

int  RenderScene::getAnimLoop()
{
    return this->mAnimationLoop ? 1 : 0;
}

double RenderScene::getAnimDuration()
{
    return this->mDuration;
}

int  RenderScene::isAnimPlay()
{
    return this->mAnimationPlay ? 1 : 0;
}
